import warnings

warnings.simplefilter('default')
warnings.warn (
   "aquests.lib will be deprecated, use rs4",
    DeprecationWarning
)

from rs4 import (
    attrdict,
    confparse,
    logger,
    pathtool,
    siesta,
    termcolor
)



import re, sys, os
try:
    from urllib.parse import unquote, unquote_plus
except ImportError:
    from urlparse import unquote
    from urllib import unquote_plus    
import json
from hashlib import md5
import random

####################################################################################
# Utitlities
####################################################################################
REQUEST = re.compile ('([^ ]+) ([^ ]+)(( HTTP/([0-9.]+))$|$)')
CONNECTION = re.compile ('Connection: (.*)', re.IGNORECASE)

def crack_query (r):
    if type (r) is bytes:
        r = r.decode ("utf8")

    if not r: return {}
    if r[0]=='?': r=r[1:]    
    arg={}
    q = [x.split('=', 1) for x in r.split('&')]
    
    for each in q:
        k = unquote_plus (each [0])
        try: 
            t, k = k.split (":", 1)
        except ValueError:
            t = "str"

        v = unquote_plus (each [1])
        if len (each) == 2:
            if t == "str":
                pass
            elif t in ("i", "int"):
                v = int (v)
            elif t in ("n", "null"):
                v = None    
            elif t in ("f", "float"):
                v = float (v)
            elif t in ("l", "list"):
                v = v.split (",")
            elif t in ("b", "bool"):
                v = v.lower () in ("1", "true", "yes")
                
        if k in arg:
            if type (arg [k]) is not type ([]):
                arg[k] = [arg[k]]
            arg[k].append (v)
            
        else:
            arg[k] = v
            
    return arg

def crack_request (r):
    m = REQUEST.match (r)
    if m.end() == len(r):        
        uri=m.group(2)        
        if m.group(3):
            version = m.group(5)
        else:
            version = None    
        return m.group(1).lower(), uri, version
    else:
        return None, None, None

def join_headers (headers):
    r = []
    for i in range(len(headers)):
        if headers[i][0] in ' \t':    
            r[-1] = r[-1] + headers[i][1:]
        else:
            r.append (headers[i])
    return r

def get_header (head_reg, lines, group=1):
    for line in lines:
        m = head_reg.match (line)
        if m and m.end() == len(line):
            return m.group (group)
    return ''

def get_header_match (head_reg, lines):
    for line in lines:
        m = head_reg.match (line)
        if m and m.end() == len(line):
            return m
    return ''        

def get_extension (path):
    dirsep = path.rfind ('/')
    dotsep = path.rfind ('.')
    if dotsep > dirsep:
        return path[dotsep+1:]
    else:
        return ''

ALNUM = '0123456789abcdefghijklmnopqrstuvwxyz'
def md5uniqid (length = 13):    
    global ALNUM
    _id = ''
    for i in range (0, length):
        _id += random.choice(ALNUM)
    return md5 (_id.encode ("utf8")).hexdigest ()[length:]

def parse_params (header_val):
    d = {}
    header_filed = ""
    last_key = None
    for each in header_val.split (";"):
        if last_key:
            d [last_key] += ";" + each
            if each [-1] == '"':
                last_key = None
            continue
        
        each = each.strip ()
        if not each: continue
        
        try:
            a, b = each.split ("=", 1)
        except ValueError:
            if header_filed:
                header_filed += ":" + each
            else:    
                header_filed = each
            continue
            
        key = a.lower ()
        d [key] = b
        if len (b) >= 2 and  b [0] == '"' and b [-1] != '"':
            last_key = key    
    
    return header_filed, d



#-------------------------------------------------------
# Asyn PostgresSQL Dispatcher
# Hans Roh (hansroh@gmail.com)
# 2015.6.9
#-------------------------------------------------------
		
DEBUG = False

try:
	import psycopg2
	
except ImportError:	
	class AsynConnect:
		def __init__ (self, address, params = None, lock = None, logger = None):
			logger ("[warn] cannot import psycopg2")
			raise ImportError ("cannot import psycopg2")

else:	
	import asyncore
	from . import dbconnect

	from psycopg2.extensions import POLL_OK, POLL_READ, POLL_WRITE
	_STATE_OK = (POLL_OK, POLL_WRITE, POLL_READ)
	_STATE_RETRY = -1
	_STATE_IGNORE = -2
	REREY_TEST = False	

	class AsynConnect (dbconnect.AsynDBConnect, asyncore.dispatcher):
		def __init__ (self, address, params = None, lock = None, logger = None):
			dbconnect.AsynDBConnect.__init__ (self, address, params, lock, logger)			
			self.cur = None
			self.retries = 0
			asyncore.dispatcher.__init__ (self)

		def retry (self):
			if self.request is None:
				return
			self.retries += 1	
			self.logger ("[warn] closed psycopg2 connection, retrying...")
			self.disconnect ()
			request, self.request = self.request, None
			self.execute (request)
			return _STATE_RETRY

		def check_state (self, state):
			if state == _STATE_RETRY:				
				return
			if state not in (_STATE_OK):
				self.logger ("[warn] psycopg2.poll() returned %s" % state)
				self.handle_close ()
			
		def poll (self):
			# 2 cases, if on requesting raise immediatly, else handle silently
			try:
				if REREY_TEST and self.writable () and self.request.retry_count == 0:
					#raise psycopg2.InterfaceError
					self.disconnect ()
				return self.socket.poll ()
			except (psycopg2.OperationalError, psycopg2.InterfaceError):
				if self.request:
					if self.request.retry_count == 0:
						self.request.retry_count += 1
						return self.retry ()
					else:
						raise
				self.logger.trace ()				
			except:
				# else usually timeout	
				if self.request:
					raise
				self.logger.trace ()
			return _STATE_IGNORE
			
		def writable (self):			
			return self.out_buffer or not self.connected
			
		def readable (self):
			return self.connected and not self.out_buffer
		
		def add_channel (self, map = None):
			return asyncore.dispatcher.add_channel (self, map)
				
		def del_channel (self, map=None):
			fd = self._fileno
			if map is None:
				map = self._map
			if fd in map:
				del map[fd]
		
		def handle_expt_event (self):
			self.handle_expt ()
			
		def handle_connect_event (self):
			state = self.poll ()			
			if state == POLL_OK:	
				self.handle_connect ()
				self.connected = True
				self.connecting = False		
			else:				
				self.check_state (state)
		
		def handle_write_event (self):		
			if not self.connected:
				self.handle_connect_event ()
			else:	
				self.handle_write ()
		
		def handle_expt (self):
			self.handle_close (psycopg2.OperationalError ("Socket Panic"))
			
		def handle_connect (self):			
			self.create_cursor ()
			
		def handle_read (self):
			state = self.poll ()			
			if self.cur and state == POLL_OK:
				self.set_event_time ()
				self.has_result = True
				self.close_case_with_end_tran ()
			else:
				self.check_state (state)
				
		def handle_write (self):
			state = self.poll ()
			if state == POLL_OK:
				if not self.cur:
					self.create_cursor ()				
				self.set_event_time ()
				self.cur.execute (self.out_buffer)
				self.out_buffer = ""
			else:				
				self.check_state (state)
				
		#-----------------------------------
		# Overriden
		#-----------------------------------
		def create_cursor (self):
			if self.cur is None:
				try:
					self.cur = self.socket.cursor ()						
				except:
					self.handle_error ()

		def close_cursor (self):
			if self.cur:				
				try:
					self.cur.close ()							
				except:
					self.logger.trace ()
				self.cur = None	

		def close_case (self):
			if self.request:
				description, data = self.cur and self.cur.description or None, None
				if description:
					try:
						data = self.fetchall ()
					except:
						self.logger.trace ()
						self.expt = asyncore.compact_traceback () [2]						
						data = None
				self.request.handle_result (description or None, self.expt, data)					
				self.request = None				
			self.close_cursor ()
			self.set_active (False)
		
		def fetchall (self):
			data = self.cur.fetchall ()				
			self.result = False
			return data

		def close (self, deactive = 1):					
			self.close_cursor ()
			asyncore.dispatcher.close (self)			
			dbconnect.AsynDBConnect.close (self, deactive)			
			
		def connect (self, force = 0):
			host, port = self.address		
			sock = psycopg2.connect (
				dbname = self.dbname,
				user = self.user,
				password = self.password,
				host = host,
				port = port,
				async_ = 1
			)			
			self.set_socket (sock)			
		
		def end_tran (self):			
			pass
		
		def _compile (self, request):
			sql = request.params [0]
			if isinstance (sql, (list, tuple)):
				sql = ";\n".join (map (str, sql)) + ";"
			try:
				sql = sql.strip ()
			except AttributeError:				
				raise dbconnect.SQLError ("Invalid SQL")
			if not sql:
				raise dbconnect.SQLError ("Empty SQL")
			return sql
		
		def execute (self, request):		
			dbconnect.DBConnect.begin_tran (self, request)
			self.out_buffer = self._compile (request)
			if not self.connected and not self.connecting:				
				self.connect ()
			else:
				state = self.poll ()
				if state != POLL_OK:
					self.reconnect ()
				else:
					self.create_cursor ()				

	


import threading, time, sys
from skitai import lifetime
import threading
try:
	import queue
except ImportError:
	import Queue as queue
	
			
class request_thread (threading.Thread):
	def __init__ (self, queue, logger, id = 0):
		threading.Thread.__init__ (self)
		self.queue = queue
		self.id = id
		self.logger = logger
		self.setDaemon (1)
		self.idle = 1
		self.command = ''		
		self.exec_time = 0.
		self.lock = threading.Lock ()
		
	def run (self):
		while 1:		
			job = self.queue.get ()		
			if job is None: break			
			with self.lock:
				self.idle = 0
				self.command = str (job)
			
			start_time = time.time()			
			try:
				job ()
			except MemoryError:
				self.logger.trace ("thread #%d" % self.id)				
				lifetime.shutdown (1, 1.0)				
			except:
				self.logger.trace ("thread #%d" % self.id)
				
			exc_time = time.time() - start_time
			with self.lock:
				self.exec_time = exc_time
				self.idle = 1				
			del job
			
	def getId (self):
		return self.id

	def status (self):
		self.lock.acquire ()
		idle = self.idle
		command = self.command
		exec_time = self.exec_time
		self.lock.release ()
		
		return {
			"object_id": id (self),
			"thread_id": self.id,
			"status": idle and "IDLE" or "BUSY", 
			"command":	command,
			"exec_time": "%d ms" % (exec_time * 1000,)
		}


class thread_pool:	
	def __init__ (self, queue, child, logger):
		self.tpool = {}
		for i in range (child):
			d = request_thread (queue, logger, i)
			self.tpool [i] = d
			d.start ()
			
	def __len__ (self):
		return len (self.tpool)
		
	def __getitem__ (self, k):		
		return self.tpool [k]
		
	def cleanup (self):
		pass
				
	def status (self):		
		d = {}
		for child_id, child in list(self.tpool.items()):
			d [child_id] = child.status ()
		return d
			

class request_queue:
	def __init__ (self):
		self.mon = threading.RLock()
		self.cv = threading.Condition (self.mon)
		self.queue = []

	def put(self, item, prior = 0):		
		with self.cv:
			if prior:
				self.queue.insert (0, item)
			else:
				self.queue.append (item)
			self.cv.notify ()		
		
	def get (self):
		with self.cv:
			while not self.queue:
				self.cv.wait()
			item = self.queue.pop(0)		
		return item
	
	def cleanup (self):
		with self.cv:
			self.queue = []		
	
	def status(self):
		return {
			"qsize": len (self.queue), 
			"in_queue": [str(x [1]) for x in self.queue]
		}	


class request_queue2 (queue.Queue):
	def __init__(self, maxsize=0):
		queue.Queue.__init__ (self, maxsize)
		self.maxq = 0
		
	def cleanup (self):
		while 1:
			try: item = self.get_nowait ()
			except queue.Empty: break
			del item
		
		for i in range (1024):
			self.put (None)
		
		while 1:
			try: self.get_nowait ()
			except queue.Empty: break			
	
	def put(self, item, block=True, timeout=None):
		qsize = self.qsize ()
		with self.mutex:
			if qsize > self.maxq:
				self.maxq = qsize		
		queue.Queue.put (self, item, block, timeout)		
			
	def status(self):
		with self.mutex:
			qsize = len (self.queue)
			maxq = self.maxq
		
		return {
			"qsize": qsize,
			"max": maxq
		}



import asynchat
import socket
import time
from .pydns import Base, Type, Class, Lib, Opcode

class UDPClient (asynchat.async_chat):
    protocol = "udp"
    zombie_timeout = 1200
    ac_in_buffer_size = 512
    
    def __init__ (self, addr, logger):
        self.addr = addr
        self.logger = logger                
        self.event_time = time.time ()
        self.creation_time = time.time ()        
        self.closed = False
        self._timeouted = False
        self.requests = {}
        self.callback = None
        asynchat.async_chat.__init__ (self)
                            
    def query (self, request, args, callback):    
        if self.callback is None:
            self.callback = callback
        self.event_time = time.time ()        
        self.header = None
        
        self.requests [request [:2]] = [args, time.time ()]
        qname = args ["name"].decode ("utf8")
        args ['addr'] = self.addr
        self.push (request)
            
        if not self.connected:
            self.set_terminator (None)
            self.create_socket (socket.AF_INET, socket.SOCK_DGRAM)
            self._connect ()
                
    def _connect (self):        
        try:
            self.connect (self.addr)
        except:
            self.handle_error ()
                    
    def trace (self):
        self.logger.trace ()
    
    def log_info (self, line, level = 'info'):
        self.log ("[%s] %s" % (level, line))
              
    def log (self, line):
        self.logger (line)
            
    def handle_error (self):
        self.trace ()
        self.close ()
    
    def handle_timeout (self):
        self._timeouted = True
        self.handle_close ()
                    
    def handle_connect (self):    
        self.event_time = time.time ()        
        
    def handle_expt (self):
        self.handle_close ()
        
    def collect_incoming_data (self, data, id = None):
        try:
            args, starttime = self.requests.pop (id or data [:2])                
        except KeyError:
            # alerady timeouted
            pass
        else:    
            self.callback (args, data, id is not None)
            
    def handle_close (self):    
        for args, starttime in self.requests.values ():
            self.callback (args, b'', True)
        self.close ()

        
class TCPClient (UDPClient):
    protocol = "tcp"
    zombie_timeout = 10
    ac_in_buffer_size = 65536
    
    def __init__ (self, addr, logger):        
        UDPClient.__init__ (self, addr, logger)
        self.callback = None
        self.args = None
        self.request = None
        self.header = None
        self.reply = b''
                            
    def query (self, request, args, callback):    
        self.event_time = time.time ()
        self.reply = b""
        self.header = None
        self.request = request
        self.args = args
        self.callback = callback
        
        args ['addr'] = self.addr        
        self.push (Lib.pack16bit(len(request)) + request)
        if not self.connected:
            self.set_terminator (2)
            self.create_socket (socket.AF_INET, socket.SOCK_STREAM)                
            self._connect ()
        
    def collect_incoming_data (self, data):        
        self.reply += data
            
    def found_terminator (self):    
        if self.header:
            self.callback (self.args, self.header + self.reply, self._timeouted)
            self.close ()
            
        else:
            self.header, self.reply = self.reply, b""
            count = Lib.unpack16bit(self.header)
            self.set_terminator (count)            
        
    def handle_close (self):    
        self.callback (self.args, b'', self._timeouted)                
        self.close ()



import sqlite3
from . import dbconnect, asynpsycopg2
import threading

DEBUG = False

class OperationalError (Exception):
	pass

class SynConnect (asynpsycopg2.AsynConnect, dbconnect.DBConnect):
	def __init__ (self, address, params = None, lock = None, logger = None):
		dbconnect.DBConnect.__init__ (self, address, params, lock, logger)
		self.connected = False
		self.conn = None
		self.cur = None		
	
	def close_if_over_keep_live (self):
		# doesn't need disconnect with local file
		pass
			
	def is_channel_in_map (self, map = None):
		return False

	def del_channel (self, map=None):
		pass
			
	def close (self, deactive = 1):	
		if self.cur:
			self.cur.close ()
			self.cur = None
		if self.conn:	
			self.conn.close ()			
			self.conn = None	
		self.connected = False	
		dbconnect.DBConnect.close (self, deactive)
				
	def close_case (self):
		asynpsycopg2.AsynConnect.close_case (self)
				
	def connect (self):
		try:
			self.conn = sqlite3.connect (self.address, check_same_thread = False)			
		except:
			self.handle_error ()
		else:	
			self.connected = True
						
	def execute (self, request):
		dbconnect.DBConnect.begin_tran (self, request)			
		sql = self._compile (request)
		
		if not self.connected:
			self.connect ()
			self.conn.isolation_level = None
				
		try:
			if self.cur is None:
				self.cur = self.conn.cursor ()

			if len (request.params) > 1 or sql [:7].lower () == "select ":
				self.cur.execute (sql, *request.params [1:])
				self.has_result = True
			else:			
				self.cur.executescript (sql)
		except:
			self.handle_error ()
		else:			
			self.close_case ()
		